/*
 * Copyright (C) 2009 Igalia S.L.
 *
 * Contact: Iago Toral Quiroga <itoral@igalia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include "mafw-simple-renderer.h"

#include <string.h>

#define MAFW_SIMPLE_RENDERER_GET_PRIVATE(object) \
        (G_TYPE_INSTANCE_GET_PRIVATE((object), MAFW_TYPE_SIMPLE_RENDERER, \
				     MafwSimpleRendererPrivate))

struct _MafwSimpleRendererPrivate {
	GstElement *pipeline;
	MafwPlaylist *pls;
	gchar *objectid;
	gint index;
	MafwPlayState state;
};

static gboolean
mafw_simple_renderer_assign_playlist (MafwRenderer *renderer,
				      MafwPlaylist *pls,
				      GError **error);

static void 
mafw_simple_renderer_play (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data);
static void
mafw_simple_renderer_stop (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data);
static void 
mafw_simple_renderer_pause (MafwRenderer *self,
			    MafwRendererPlaybackCB callback,
			    gpointer user_data);
static void 
mafw_simple_renderer_resume (MafwRenderer *self,
			     MafwRendererPlaybackCB callback,
			     gpointer user_data);

static void
mafw_simple_renderer_get_status (MafwRenderer *self,
				 MafwRendererStatusCB callback,
				 gpointer user_data);
static void
mafw_simple_renderer_next (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data);
static void
mafw_simple_renderer_prev (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data);
static void
mafw_simple_renderer_goto_index (MafwRenderer *self,
				 guint index,
				 MafwRendererPlaybackCB callback,
				 gpointer user_data);
static void
mafw_simple_renderer_get_position (MafwRenderer *self,
				   MafwRendererPositionCB callback,
				   gpointer user_data);
static void
mafw_simple_renderer_set_position (MafwRenderer *self,
				   MafwRendererSeekMode mode,
				   gint seconds,
				   MafwRendererPositionCB callback,
				   gpointer user_data);

static void playback_play (MafwSimpleRenderer *renderer, GError **error);
static void playback_stop (MafwSimpleRenderer *renderer);
static void playback_resume (MafwSimpleRenderer *renderer);
static void playback_pause (MafwSimpleRenderer *renderer);

/*________________________ Plugin init  ________________________*/

/*
 * Plugin descriptor
 */
G_MODULE_EXPORT MafwPluginDescriptor mafw_simple_renderer_plugin_description =
{
	{ .name = MAFW_SIMPLE_RENDERER_PLUGIN_NAME },
	.initialize = mafw_simple_renderer_plugin_initialize,
	.deinitialize = mafw_simple_renderer_plugin_deinitialize,
};

/*
 * Plugin initialization. Creates an instance of the renderer and 
 * registers it so it becomes available for applications. A plugin
 * could spawn more than one renderer instance, and each one must have
 * a unique UUID (see mafw-upnp-source for an example of this).
 */
gboolean
mafw_simple_renderer_plugin_initialize (MafwRegistry *registry, GError **error)
{
	MafwRenderer *renderer;

	/* Init GStreamer */
	gst_init (NULL, NULL);
	gst_pb_utils_init ();

	/* Create a simple renderer instance and register it */
	renderer = MAFW_RENDERER (
		g_object_new (MAFW_TYPE_SIMPLE_RENDERER,
			      "plugin", MAFW_SIMPLE_RENDERER_PLUGIN_NAME,
			      "uuid", MAFW_SIMPLE_RENDERER_UUID,
			      "name", MAFW_SIMPLE_RENDERER_NAME,
			      NULL));

	mafw_registry_add_extension (registry, MAFW_EXTENSION (renderer));
	
	return TRUE;
}

/*
 * Plugin deinit, here we should free any resources used by the plugin.
 */
void 
mafw_simple_renderer_plugin_deinitialize (GError **error) 
{
}


/*_____________________ GObject setup ________________________ */

G_DEFINE_TYPE (MafwSimpleRenderer, mafw_simple_renderer, MAFW_TYPE_RENDERER);

static void
mafw_simple_renderer_class_init (MafwSimpleRendererClass *klass)
{
	MafwRendererClass *renderer_class = MAFW_RENDERER_CLASS (klass);

	/* We only implement a subset of the MafwRenderer's API */
	renderer_class->play = mafw_simple_renderer_play;
	renderer_class->pause = mafw_simple_renderer_pause;
	renderer_class->resume = mafw_simple_renderer_resume;
	renderer_class->stop = mafw_simple_renderer_stop;
	renderer_class->get_status = mafw_simple_renderer_get_status;
	renderer_class->assign_playlist = mafw_simple_renderer_assign_playlist;
	renderer_class->next = mafw_simple_renderer_next;
	renderer_class->previous = mafw_simple_renderer_prev;
	renderer_class->goto_index = mafw_simple_renderer_goto_index;
	renderer_class->set_position = mafw_simple_renderer_set_position;
	renderer_class->get_position = mafw_simple_renderer_get_position;
/* 	renderer_class->get_current_metadata = mafw_simple_renderer_get_meta; */

	g_type_class_add_private(klass, sizeof (MafwSimpleRendererPrivate));
}

static void
mafw_simple_renderer_init (MafwSimpleRenderer *renderer)
{
	renderer->priv = MAFW_SIMPLE_RENDERER_GET_PRIVATE (renderer);
	renderer->priv->pls = NULL;
	renderer->priv->pipeline = NULL;
	renderer->priv->objectid = NULL;
	renderer->priv->index = 0;
	renderer->priv->state = Stopped;
}

/*__________________________ Utilities ____________________________ */

static void
handle_error (MafwSimpleRenderer *renderer, GError *error)
{
	/* Just signal the error to the clients and get back
	   to Stopped state if we were playing */
	g_signal_emit_by_name (renderer,
			       "error",
			       MAFW_RENDERER_ERROR,
			       MAFW_EXTENSION_ERROR_FAILED,
			       error->message);
	playback_stop (renderer);
}

static void
set_state (MafwSimpleRenderer *renderer, MafwPlayState state)
{
	/* Store the current state and signal the state change
	   to clients */
	renderer->priv->state = state;
	g_signal_emit_by_name (MAFW_RENDERER (renderer),
			       "state-changed", state);
}

static void
select_media (MafwSimpleRenderer *renderer, guint index, GError **error)
{
	renderer->priv->index = index;

	/* Get the objectid of the item at 'index' */
	renderer->priv->objectid = 
		mafw_playlist_get_item (renderer->priv->pls,
					renderer->priv->index,
					error);

	/* And emit the new media selected */
	if (error != NULL) {
		g_signal_emit_by_name (MAFW_RENDERER (renderer),
				       "media-changed",
				       renderer->priv->index,
				       renderer->priv->objectid);
	}
}

static void
handle_eos (MafwSimpleRenderer *renderer)
{
	guint size = 0;
	GError *error = NULL;
	gboolean repeat = FALSE;

	/* Move to the next in the playlist automatically */
	size = mafw_playlist_get_size (renderer->priv->pls, &error);
	
	if (error) {
		handle_error (renderer, error);
		g_error_free (error);
	} else if (renderer->priv->index < size - 1) {
		select_media (renderer, renderer->priv->index + 1, &error);
		if (error == NULL) {
			playback_play (renderer, &error);
		}
		if (error != NULL) {
			handle_error (renderer, error);
			g_error_free (error);
		}
	} else {
		/* We reached the end of the playlist, check if we
		   are in repeat mode, and in that case restart */
		repeat = mafw_playlist_get_repeat (renderer->priv->pls);
		if (repeat) {
			select_media (renderer, 0, &error);
			if (error == NULL) {
				playback_play (renderer, &error);
			}
			if (error != NULL) {
				handle_error (renderer, error);
				g_error_free (error);
			}
		} else {
			playback_stop (renderer);
		}
	}
}

static void
handle_state_changed (MafwSimpleRenderer *renderer, GstStateChange state)
{
	switch (state) {
	case GST_STATE_CHANGE_READY_TO_NULL:
		set_state (renderer, Stopped);
		break;
	case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
		set_state (renderer, Playing);
		break;
	case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
		set_state (renderer, Paused);
		break;
	default:
		break;
	}
}

static gboolean
gst_bus_handler (GstBus *bus, GstMessage *msg, MafwSimpleRenderer *renderer)
{
	GstState old, new;
	GstStateChange state_change;
	GError *error = NULL;

	switch (GST_MESSAGE_TYPE (msg))
	{
	case GST_MESSAGE_ERROR:
		gst_message_parse_error (msg, &error, NULL);
		handle_error (renderer, error);
		g_error_free (error);
		break;
	case GST_MESSAGE_EOS:
		handle_eos (renderer);
		break;
	case GST_MESSAGE_STATE_CHANGED:
		gst_message_parse_state_changed (msg, &old, &new, NULL);
		state_change = GST_STATE_TRANSITION (old, new);
		handle_state_changed (renderer, state_change);
		break;
	default:
		break;
	}

	return TRUE;
}

static void
play_uri (MafwSimpleRenderer *renderer, gchar *uri)
{
	GstBus *bus = NULL;

	playback_stop (renderer);

	/* Create the pipeline */
	renderer->priv->pipeline = 
		gst_element_factory_make ("playbin2", "pipeline");
	if (renderer->priv->pipeline == NULL) {
		g_signal_emit_by_name (renderer,
				       "error",
				       MAFW_RENDERER_ERROR,
				       MAFW_RENDERER_ERROR_UNABLE_TO_PERFORM,
				       "Failed to create pipeline");
		return;
	}
	
	bus = gst_pipeline_get_bus (GST_PIPELINE (renderer->priv->pipeline));
	gst_bus_add_watch_full (bus,
				G_PRIORITY_HIGH,
				(GstBusFunc) gst_bus_handler,
				renderer, 
				NULL);

	/* Provide the URI and play */
	g_object_set (G_OBJECT (renderer->priv->pipeline), "uri", uri, NULL);
	gst_element_set_state (renderer->priv->pipeline, GST_STATE_PLAYING);
}

static void
metadata_cb (MafwSource *source,
	     const gchar *objectid,
	     GHashTable *metadata,
	     gpointer user_data,
	     const GError *error)
{
	GValue *value = NULL;
	gchar *uri = NULL;
	MafwSimpleRenderer *renderer = NULL;

	renderer = MAFW_SIMPLE_RENDERER (user_data);

	/* Check if we failed to resolve the URI */
	if (error != NULL) {
		g_critical ("Failed to get URI from source, we cannot play");
		/* We notify os async. error via this signal */
		g_signal_emit_by_name (renderer, 
				       "error", 
				       error->domain, 
				       error->code, 
				       error->message);
		return;
	}

	/* Maybe we stopped playback before we got the URI, in that
	   case we just ignore this */
	if (!objectid || !renderer->priv->objectid ||
	    strcmp (objectid, renderer->priv->objectid)) {
		return;
	}

	/* Extract the URI from the result */
	value = mafw_metadata_first (metadata, MAFW_METADATA_KEY_URI);
	uri = value ? (gchar *) g_value_get_string (value) : NULL;
	if (uri == NULL) {
		g_critical ("Failed to get URI from source, we canot play");
		g_signal_emit_by_name (renderer, 
				       "error", 
				       error->domain, 
				       error->code, 
				       error->message);
	}

	/* So we got a valid URI, time to play! */
	play_uri (renderer, uri);
}

static void
playback_play (MafwSimpleRenderer *renderer, GError **error)
{
	MafwSource *source = NULL;
	MafwRegistry *registry = NULL;
	gchar *source_uuid;
	const gchar *const *keys = NULL;

	if (!renderer->priv->objectid) {
		if (error != NULL)
			*error = g_error_new (MAFW_RENDERER_ERROR,
					      MAFW_RENDERER_ERROR_UNABLE_TO_PERFORM,
					      "No objectid to play");
		return;
	}

	/* Now we need to get the URI for this objectid so we can
	   feed it to GStreamer. For that, we first get the source UUID
	   so we know which source we need to query */
	mafw_source_split_objectid (renderer->priv->objectid, 
				    &source_uuid, NULL);
	registry = MAFW_REGISTRY (mafw_registry_get_instance ());
	source = MAFW_SOURCE (mafw_registry_get_extension_by_uuid(
				      registry, source_uuid));
	g_free (source_uuid);
	
	if (source == NULL) {
		if (error != NULL)
			*error = g_error_new (MAFW_RENDERER_ERROR,
					      MAFW_RENDERER_ERROR_MEDIA_NOT_FOUND,
					      "Could not identify media source");
		return;
	}

	/* And once we located the source, we query it for the URI */
	keys = MAFW_SOURCE_LIST (MAFW_METADATA_KEY_URI);
	mafw_source_get_metadata (source,
				  renderer->priv->objectid,
				  keys,
				  metadata_cb,
				  renderer);
}

static void
playback_stop (MafwSimpleRenderer *renderer)
{
	gst_element_set_state (renderer->priv->pipeline, GST_STATE_NULL);
	gst_object_unref (GST_OBJECT (renderer->priv->pipeline));
	renderer->priv->pipeline = NULL;
	renderer->priv->state = Stopped;
	g_free (renderer->priv->objectid);
	renderer->priv->objectid = NULL;
	set_state (renderer, Stopped);
}

static void
playback_pause (MafwSimpleRenderer *renderer)
{
	gst_element_set_state (renderer->priv->pipeline, GST_STATE_PAUSED);
}

static void
playback_resume (MafwSimpleRenderer *renderer)
{
	gst_element_set_state (renderer->priv->pipeline, GST_STATE_PLAYING);
}

/*_____________________ API implementation ________________________ */

static gboolean
mafw_simple_renderer_assign_playlist (MafwRenderer *self,
				      MafwPlaylist *pls,
				      GError **error)
{
	MafwSimpleRenderer *renderer = NULL;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));
	renderer = MAFW_SIMPLE_RENDERER (self);

	/* If we had a playlist assigned already, stop playback and
	   release the playlist */
	if (renderer->priv->pls != NULL) {
		playback_stop (renderer);
		mafw_playlist_decrement_use_count (renderer->priv->pls, NULL);
		g_object_unref (renderer->priv->pls);
		renderer->priv->pls = NULL;
	}

	/* Assign the new playlist */
	if (pls != NULL) {
		renderer->priv->pls = g_object_ref (pls);
		mafw_playlist_increment_use_count (renderer->priv->pls, NULL);
		/* Probably we want to connect to playlist::contents-changed
		   signal here */
	}

	/* Get ready to play this new playlist from the beginning */
	select_media (renderer, 0, error);

	/* Notify that we have a new playlist assigned */
	g_signal_emit_by_name (MAFW_RENDERER (renderer), 
			       "playlist-changed",
			       renderer->priv->pls);

	return TRUE;
}

static void 
mafw_simple_renderer_play (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data)
{
	GError *error = NULL;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));	
	playback_play (MAFW_SIMPLE_RENDERER (self), &error);
	callback (self, user_data, error);
	if (error)
		g_error_free (error);
}

static void
mafw_simple_renderer_stop (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data)
{
	g_assert (MAFW_IS_SIMPLE_RENDERER (self));	
	playback_stop (MAFW_SIMPLE_RENDERER (self));
	callback (self, user_data, NULL);
}

static void 
mafw_simple_renderer_pause (MafwRenderer *self,
			    MafwRendererPlaybackCB callback,
			    gpointer user_data)
{
	g_assert (MAFW_IS_SIMPLE_RENDERER (self));	
	playback_pause (MAFW_SIMPLE_RENDERER (self));
	callback (self, user_data, NULL);
}

static void 
mafw_simple_renderer_resume (MafwRenderer *self,
			     MafwRendererPlaybackCB callback,
			     gpointer user_data)
{
	g_assert (MAFW_IS_SIMPLE_RENDERER (self));	
	playback_resume (MAFW_SIMPLE_RENDERER (self));
	callback (self, user_data, NULL);
}

static void
mafw_simple_renderer_get_status (MafwRenderer *self,
				 MafwRendererStatusCB callback,
				 gpointer user_data)
{
	MafwSimpleRenderer *renderer = NULL;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));	
	renderer = MAFW_SIMPLE_RENDERER (self);

	callback (self,
		  renderer->priv->pls,
		  renderer->priv->pls ? renderer->priv->index : -1,
		  renderer->priv->state,
		  renderer->priv->objectid,
		  user_data,
		  NULL);
}
			      
static void
mafw_simple_renderer_next (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data)
{
	GError *error = NULL;
	guint size = 0;
	MafwSimpleRenderer *renderer = NULL;
	gboolean repeat = FALSE;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));	
	renderer = MAFW_SIMPLE_RENDERER (self);
	g_assert (renderer->priv->pls);

	/* First we get the size of the playlist */
	size = mafw_playlist_get_size (renderer->priv->pls, &error);	
	if (error != NULL) {
		goto done;
	}

	/* And then we check if we have reached the end */
	if (renderer->priv->index < size - 1) {
		select_media (renderer, renderer->priv->index + 1, &error);
	} else {
		/* We reached the end of the playlist, check if we
		   are in repeat mode, and in that case restart */
		repeat = mafw_playlist_get_repeat (renderer->priv->pls);
		if (repeat) {
			select_media (renderer, 0, &error);
		} else {
			playback_stop (renderer);
		}
	}

	/* If we moved to the next successfully and we were not
	   stopped, start playing */
	if (error == NULL && renderer->priv->state == Playing) {
		playback_play (renderer, &error);
	}

done:
        callback (self, user_data, error);
        if (error)
		g_error_free (error);
}

static void
mafw_simple_renderer_prev (MafwRenderer *self,
			   MafwRendererPlaybackCB callback,
			   gpointer user_data)
{
	GError *error = NULL;
	guint size = 0;
	MafwSimpleRenderer *renderer = NULL;
	gboolean repeat = FALSE;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));	
	renderer = MAFW_SIMPLE_RENDERER (self);
	g_assert (renderer->priv->pls);

	/* Check if we are already at the head of the playlist */
	if (renderer->priv->index > 0) {
		select_media (renderer, renderer->priv->index - 1, &error);
	} else {
		/* We are already at the head, check if we
		   are in repeat mode, and in that case go to
		   the tail */
		repeat = mafw_playlist_get_repeat (renderer->priv->pls);
		if (repeat) {
			size = mafw_playlist_get_size (renderer->priv->pls,
						       &error);	
			if (!error) {
				select_media (renderer, size - 1, &error);
			}
		} else {
			playback_stop (renderer);
		}
	}

	/* We moved to the prev, now if we were not stopped,
	   start playing */
	if (error == NULL && renderer->priv->state == Playing) {
		playback_play (renderer, &error);
	}

        callback (self, user_data, error);
        if (error)
		g_error_free (error);
}

static void
mafw_simple_renderer_goto_index (MafwRenderer *self,
				 guint index,
				 MafwRendererPlaybackCB callback,
				 gpointer user_data)
{
	MafwSimpleRenderer *renderer = NULL;
	guint size = 0;
	GError *error = NULL;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));
	renderer = MAFW_SIMPLE_RENDERER (self);
	g_assert (renderer->priv->pls);

	/* First, check if the index is valid */
	size = mafw_playlist_get_size (renderer->priv->pls, &error);	
	if (error) {
		goto done;
	}

	if (index > size - 1) {
		error = g_error_new (MAFW_RENDERER_ERROR,
				     MAFW_PLAYLIST_ERROR_INVALID_INDEX,
				     "Index out of bounds");
		goto done;
	}

	/* It is, so move to it */
	select_media (renderer, index, &error);

	/* If we were not stopped, start playing */
	if (error == NULL && renderer->priv->state == Playing) {
		playback_play (renderer, &error);
	}

done:
        callback (self, user_data, error);
        if (error)
		g_error_free (error);
}

static void
mafw_simple_renderer_get_position (MafwRenderer *self,
				   MafwRendererPositionCB callback,
				   gpointer user_data)
{
	MafwSimpleRenderer *renderer = NULL;
	gint64 position = 0;
	GstFormat format = GST_FORMAT_TIME;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));
	renderer = MAFW_SIMPLE_RENDERER (self);

	if (renderer->priv->state != Stopped) {
		gst_element_query_position (renderer->priv->pipeline,
					    &format, &position);
	}

	callback (self, GST_TIME_AS_SECONDS (position), user_data, NULL);
}

static void
mafw_simple_renderer_set_position (MafwRenderer *self,
				   MafwRendererSeekMode mode,
				   gint position,
				   MafwRendererPositionCB callback,
				   gpointer user_data)
{
	MafwSimpleRenderer *renderer = NULL;
	gboolean res = TRUE;
	GError *error = NULL;

	g_assert (MAFW_IS_SIMPLE_RENDERER (self));
	renderer = MAFW_SIMPLE_RENDERER (self);

	/* We can only seek if we are not stopped */
	if (renderer->priv->state != Stopped) {
		res = gst_element_seek_simple (renderer->priv->pipeline,
					       GST_FORMAT_TIME,
					       GST_SEEK_FLAG_FLUSH | 
					       GST_SEEK_FLAG_KEY_UNIT,
					       (gint64) position * GST_SECOND);
		if (res == FALSE) {
			error = g_error_new (MAFW_RENDERER_ERROR,
					     MAFW_RENDERER_ERROR_UNABLE_TO_PERFORM,
					     "Seek operation failed");
		}
	}

	callback (self, position, user_data, error);	
	if (error)
		g_error_free (error);
}
